/*
 * Copyright (c) 2009-2010 by Oleg Trifonov <otrifonow@gmail.com>
 *
 * http://trol.0fees.net
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of either the GNU General Public License version 2
 * or the GNU Lesser General Public License version 2.1, both as
 * published by the Free Software Foundation.
 */

#include <cstdio>

#include "Device.h"
#include "Timer.h"

Device::Device() {
	noWrite = false;
}


Device::~Device() {
}


// Connect to the bootloader
bool Device::Open(const char *comName, int baudRate) {
	// open COM-port
	if ( !com.Open(comName, baudRate) ) {
		printf("can't open %s\n", comName);
		return false;
	}
/*
	printf("Waiting...\n");
	// it's need to wait of device initialization and getting a response 'ok' from it
	while ( com.ReadByte() != 'o' );
	if ( (com.ReadByte() != 'k') && (com.ReadByte() != 0x0D) ) {
		printf("invalid device response\n");
		return false;
	}
*/
	return true;
}


// Set the Z-register
bool Device::CommandZ(int z) {
	com.WriteByte('Z');
	sendHexWord(z);
	if ( com.ReadByte() != 0x0D )
		return false;
	return true;
}


// Reads a data block
bool Device::CommandR(unsigned char *data, int size) {
	com.WriteByte('R');
	sendHexByte(size);
	for (int i = 0; i < size; i++) {
		int b = getHexByte();
		if ( b < 0 )
			return false;
		data[i] = b;
	}
	if ( com.ReadByte() != 0x0D )
		return false;
	return true;
}


// Writes a data block
bool Device::CommandW(unsigned char *data, int size) {
	if ( size == 0 )
		return true;
	if ( size % 2 != 0 )
		return false;
	com.WriteByte('W');
	sendHexByte(size/2);
	for (int i = 0; i < size; i++)
		sendHexByte(data[i]);
	if ( com.ReadByte() != 0x0D )
		return false;
	return true;
}


// Reads the R0:R1 pair, then one byte into SPMCR and executes the SPM instruction
bool Device::CommandP(int r0r1, int spmcr) {
	com.WriteByte('P');
	sendHexWord(r0r1);
	sendHexByte(spmcr);
	if ( com.ReadByte() != 0x0D )
		return false;
	return true;
}


// Reads a one memory page
bool Device::ReadPage(int offset, unsigned char *data, int size) {
	if ( !CommandZ(offset) )
		return false;
	if ( !CommandR(data, size) ) 
		return false;
	return true;
}


// Reads all chip memory
bool Device::ReadAll(unsigned char *data, int size) {
	Timer timer;

	timer.Start();
	int offset = 0;
	for ( int page = 0; page < size/pageSize; page++ ) {
		Progress("Reading", 100*offset/size, timer.GetTime());
		if ( !ReadPage(offset, &data[offset], pageSize) )
			return false;
		offset += pageSize;
	}
	timer.Stop();
	Progress("Reading", 100, timer.GetTime());
	if ( !quellprogress )
		printf("\n");	
	return true;
}


// Writes a one memory page
bool Device::WritePage(int offset, unsigned char *data, int size) {
	if ( !CommandZ(offset) )
		return false;
	if ( !noWrite ) {
		// Erase a page
		if ( !CommandP(0x0000, 0x03) )
			return false;
		// Mark this page section as available
		if ( !CommandP(0x0000, 0x11) )
			return false;
	}

	// [???] may be it's a some delay? [???]
	if ( !GetLoaderOffset() )
		return false;

	// Send a data to the page buffer
	if ( !CommandW(data, size) )
		return false;

	if ( !CommandZ(offset) )
		return false;


	// [???] may be it's a some delay? [???]
	if ( !GetLoaderOffset() )
		return false;

	if ( !noWrite ) {
		// Write a page
		if ( !CommandP(0x0000, 0x05) )
			return false;
		// Mark this page section as available
		if ( !CommandP(0x0000, 0x11) )
			return false;
	}

	// [???] may be it's a some delay? [???]
	if ( !GetLoaderOffset() )
		return false;

	return true;
}


// Gets a bootloader start offset
int Device::GetLoaderOffset() {
	com.WriteByte('Q');
	getHexWord();
	int result = getHexWord();
	if ( com.ReadByte() != 0x0D )
		return false;
	return result;
}


// Writes all memory pages
// If smart == true, then before writing of a page it will be read completely.
// And if changes is not required, the page won't write
// If byte value is less than zero, this byte won't changed
bool Device::WriteAll(int *data, int size, bool smart) {
	Timer timer;

	timer.Start();
	pagesWrited = 0;
	int pagesCount = size/pageSize;
	if ( size % pageSize > 0 )
		pagesCount++;
	pagesWritedTotal = pagesCount;
	if ( pagesCount == 0 )
		return true;


	// if the "smart" mode is enabled, that read all necessary pages
	unsigned char *readed = 0;
	if ( smart ) {
		 readed = new unsigned char[pagesCount*pageSize];
		 if ( !ReadAll(readed, pagesCount*pageSize) ) {
			 delete[] readed;
			 return false;
		 }
	}
	int offset = 0;
	for (int page = 0; page < pagesCount; page++) {
		bool writePage = false;	// need to rewrite this page
		bool needReading = false;// need to read this page

		Progress("Writing", 100*page/pagesCount, timer.GetTime());
		// let's check up, whether it's necessary to update this page
		for (int i = 0; i < pageSize; i++) {
			int addr = offset + i;
			if ( addr >= size ) {
				needReading = true;
				break;
			}
			if ( data[addr] < 0 ) {
				needReading = true;
				continue;
			}
			if ( smart && data[addr] == readed[addr] ) {
				continue;
			}
			writePage = true;
		}
		
		if ( writePage ) {
			pagesWrited++;
			// prepare page data before writing
			unsigned char *pageData = new unsigned char[pageSize];
			// if a part of the pagedata is not defined, need to read it at first
			if ( needReading ) {
				if ( smart ) {
					// page data already readed
					for (int i = 0; i < pageSize; i++) 
						pageData[i] = readed[offset + i];
				} else {
					if ( !ReadPage(offset, pageData, pageSize) ) {
						delete[] pageData;
						return false;
					}
				}
			}
			// now change all necessary bytes
			for (int i = 0; i < pageSize; i++) 
				if ( offset + i < size && data[offset + i] >= 0 )
					pageData[i] = data[offset + i];

			// writes a page
			if ( !WritePage(offset, pageData, pageSize) ) {
				delete[] pageData;
				return false;
			}
			delete[] pageData;
		} // writePage

		offset += pageSize;
	} // for (page)

	if ( smart )
		delete[] readed;
	timer.Stop();
	Progress("Writing", 100, timer.GetTime());
	if ( !quellprogress )
		printf("\n");
	return true;
}


// Jump to program offset
bool Device::Jump(int offset) {
	if ( !CommandZ(offset) )
		return false;
	com.WriteByte('@');
	if ( com.ReadByte() != 0x0D )
		return false;
	return true;
}


// Reset the bootloader and verify the signature
bool Device::Reset() {
//	for (int i = 0; i < 10; i++)
//		com.WriteByte('!');
	// get the bootloader offset
	com.WriteByte('Q');
	char c1 = com.ReadByte();
	char c2 = com.ReadByte();
	if ( c1 == 'o' && c2 == 'k' ) {	// if a programmator has been started earlier than the controller
		if ( com.ReadByte() != 0x0D )
			return false;
		com.WriteByte('Q');
		if ( getHexWord() < 0 )
			return false;
	} else {	// otherwise read Z
		if ( getHexByte() < 0 )
			return false;
	}
	int bootOffset = getHexWord() + 0x0E;	// [!!!] ignore verifyication
	if ( com.ReadByte() != 0x0D )
		return false;
	if ( !CommandZ(bootOffset) )
		return false;	

	com.WriteByte('Q');
	int zz = getHexWord();
//	printf("JMP  zz = %x\n", zz);
	int bs = getHexWord();
//	printf("JMP  bs = %x\n", bs);
	if ( com.ReadByte() != 0x0D )
		return false;
/*
	com.WriteByte('@');
	if ( com.ReadByte() != 'o' )
		return false;
	if ( com.ReadByte() != 'k' )
		return false;
	if ( com.ReadByte() != 0x0D )
		return false;
*/
	return true;
}


// Sends one byte in hex format
void Device::sendHexByte(unsigned char b) {
	unsigned char hi = (b >> 4) & 0x0F;
	unsigned char lo = b & 0x0F;
	com.WriteByte(hi < 10 ? hi + '0' : hi + 'a' - 10);
	com.WriteByte(lo < 10 ? lo + '0' : lo + 'a' - 10);
}


// Sends one word in hex format
void Device::sendHexWord(unsigned short w) {
	sendHexByte((w >> 8) & 0xFF);
	sendHexByte(w & 0xFF);
}


// Reads one byte in hex format. returns -1 if there was an error
int Device::getHexByte() {
	unsigned char hi = com.ReadByte();
	unsigned char lo = com.ReadByte();
	if ( hi >= '0' && hi <= '9' ) {
		hi -= '0';
	} else if ( hi >= 'a' && hi <= 'f' ) {
		hi -= 'a' - 10;
	} else
		return -1;
	if ( lo >= '0' && lo <= '9' ) {
		lo -= '0';
	} else if ( lo >= 'a' && lo <= 'f' ) {
		lo -= 'a' - 10;
	} else
		return -1;
	return (hi << 4) + lo;
}


// Reads one byte in hex format. returns -1 if there was an error
int Device::getHexWord() {
	int hi = getHexByte();
	if ( hi < 0 )
		return -1;
	int lo = getHexByte();
	if ( lo < 0 )
		return -1;
	return (hi << 8) + lo;
}


// Prints the progressbar
void Device::Progress(const char* msg, int percent, float time) {
	if ( quellprogress )
		return;
	char progress[51];
	progress[50] = 0;
	for ( int i = 0; i < 50; i++ )
		if ( 2*i <= percent )
			progress[i] = '#';
		else
			progress[i] = ' ';
	printf("\x0D%s | %s | %i%% %.2fs", msg, progress, percent, 0.000001*time);
}