/*
 * Copyright (c) 2009-2010 by Oleg Trifonov <otrifonow@gmail.com>
 *
 * http://trol.0fees.net
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of either the GNU General Public License version 2
 * or the GNU Lesser General Public License version 2.1, both as
 * published by the Free Software Foundation.
 */

#include <cstdio> 

#include "DataFile.h"
#include "Device.h"
#include "Config.h"
#include "StrUtils.h"



void usage() {
	printf(
		"t-bootloader v1.01.  Copyright 2009-2010 Oleg Trifonov\n"
		"Usage: tboot [options]\n"
		"Options:\n"
		"   -p <partno>                Specify AVR device.\n"
		"   -b <baudrate>              Override RS-232 baud rate.\n"
		"   -C <config-file>           Specify location of configuration file.\n"
		"   -P <port>                  Specify connection port.\n"
      "   -U <flash|eeprom>:r|w|v:<filename.ext>\n"
		"                              Memory operation specification.\n"
      "                              Multiple -U options are allowed, each request\n"
      "                              is performed in the order specified.\n"
		"   -n                         Do not write anything to the device.\n"
		"   -V                         Do not verify.\n"
		"   -v                         Verbose output.\n"
		"   -q                         Quell progress output.\n"
		);
}



int main(int argc, char* argv[]) {
	Config cfg;
	Device dev;

	if ( argc == 1 ) {
		usage();
		return 1;
	}
	if ( !cfg.ParseCommandLine(argc, argv) )
		return 1;
	if ( !cfg.Load() )
		return 2;
	if ( cfg.device.length() == 0 ) {
		printf("Error: device name not defined\n");
		return 1;
	}
	if ( !cfg.SelectDevice(cfg.device) ) 
		return 1;
	
	if ( !dev.Open(cfg.port.c_str(), cfg.baudrate) )
		return 3;

	if ( !cfg.verbose ) {
		printf("Device: %s\n", cfg.device_desc.c_str());
	}

	if ( !dev.Reset() ) {
		printf("bootloader not found\n");
		return 4;
	}
	dev.SetPageSize(cfg.device_page_size);
	dev.SetOutputLevel(cfg.verbose, cfg.quellprogress);
	if ( cfg.no_write ) {
		printf("WARNING! writing is disabled\n");
		dev.DisableWriting();
	}
/*
	if (!dev.Jump(0x0000))
		printf("!!!!!!!!\n");
	else
		printf(":)\n");
*/

	// iterate for tasks
	for ( size_t task = 0; task < cfg.operations.size(); task++ ) {
		Config::TOperation *op = cfg.operations[task];
		switch ( op->type ) {
			// reading
			case 'r': {
				unsigned char *data;
				if ( op->target == 'f' ) {
					if ( !cfg.verbose )
						printf("reading flash input file \"%s\"\n", op->filename.c_str());
					data = new unsigned char[cfg.device_rom_size];
					if ( !dev.ReadAll(data, cfg.device_rom_size) ) {
						delete[] data;
						printf("chip flash reading error\n");
						return 4;
					}
				} else if ( op->target == 'e' ) {
					if ( !cfg.verbose )
						printf("reading EEPROM input file \"%s\"\n", op->filename.c_str());
					printf("EEPROM not supported in this version");
					return 10;
				}
				if ( !DataFile::SaveFile(op->filename, data, cfg.device_rom_size) ) 
					return 5;			
				delete[] data;
			   }
				break;
			// writing
			case 'w': {
				DataFile dataFile;
				if ( !cfg.verbose )
					printf("reading input file \"%s\"\n", op->filename.c_str());				 
				if ( !dataFile.Load(op->filename) ) 
					return 5;
				if ( op->target == 'f' ) {
					if ( !cfg.verbose ) {
						printf("writing flash (%i bytes)\n", dataFile.size);
						if ( cfg.smart )
							printf("smart-mode enabled\n");
					}
					// chip flash size verifycation
					if ( dataFile.size > dev.GetLoaderOffset() ) {
						printf("error! data too large: available %i bytes\n", dev.GetLoaderOffset());
						return 6;
					}
					// preparing a data array
					int *data = new int[dataFile.size];
					for ( int i = 0; i < dataFile.size; i++ )
						data[i] = dataFile.data[i];
					bool writeResult = dev.WriteAll(data, dataFile.size, cfg.smart);
					delete[] data;
					if ( !writeResult ) {
						printf("chip flash writing error\n");
						return 4;
					}
					if ( cfg.smart && !cfg.verbose )
						printf("smart mode: %i pages from %i have been rewritten (%.2f%%)\n", dev.pagesWrited, dev.pagesWritedTotal, 100.0*dev.pagesWrited/dev.pagesWritedTotal);
				} else if ( op->target == 'e' ) {
					printf("EEPROM not supported in this version");
					return 10;
				}
			 }
				if ( !cfg.verify )
					break;
			// verifying
			case 'v': {
				DataFile dataFile;
				if ( !dataFile.Load(op->filename) ) 
					return 5;
				unsigned char *readed = NULL;
				size_t readedSize = 0;

				if ( op->target == 'f' ) {
					if ( !cfg.verbose )
						printf("verifying flash against \"%s\"\n", op->filename.c_str());
					readedSize = cfg.device_rom_size;
					readed = new unsigned char[readedSize];
					if ( !dev.ReadAll(readed, cfg.device_rom_size) ) {
						delete[] readed;
						printf("chip flash reading error\n");
						return 4;
					}
				} else if ( op->target == 'e' ) {
					printf("EEPROM not supported in this version");
					return 10;
				}
				if ( dataFile.size > readedSize ) {
					printf("verifycation error: source file to large\n");
				}
				size_t len = dataFile.size < readedSize ? dataFile.size : readedSize;
				bool equals = true;
				for ( size_t i = 0; i < len; i++) {
					if ( i < readedSize && dataFile.data[i] >= 0 && dataFile.data[i] != readed[i] ) {
						printf("verifycation error: address=%s chip=%s file=%s\n", HexToStr(i, 2).c_str(), HexToStr(readed[i], 1).c_str(),HexToStr(dataFile.data[i] , 1).c_str());
						equals = false;
					}
				}				
				printf("%i bytes of %s verified\n", dataFile.definedBytes, op->target == 'f' ? "flash" : "eeprom");
				delete[] readed;
			}
				break;			
		} // switch (task)
		if ( !cfg.verbose )
			printf("\n");
	} // for ( task )



	return 0;
}

